/*
 * This file is part of CashDrainGS, which is a GameScript for OpenTTD
 * Copyright (C) 2012  Leif Linse
 *
 * CashDrainGS is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * CashDrainGS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CashDrainGS; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

/*** Libraries ***/

/* Import SuperLib for GameScript */
import("util.superlib", "SuperLib", 38);
Result <- SuperLib.Result;
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;
ScoreList <- SuperLib.ScoreList;
Tile <- SuperLib.Tile;
Direction <- SuperLib.Direction;
Town <- SuperLib.Town;
Industry <- SuperLib.Industry;
Story <- SuperLib.Story;

/* Import SCP */
import("Library.SCPLib", "SCPLib", 45);

/*** GS includes ***/
require("version.nut"); // get SELF_VERSION

require("companydata.nut");
require("scp.nut");


// Use this pointer to access the SCPManager instance that wraps up the SCP library
// with high level communication methods.
g_scp_manager <- null;


class MainClass extends GSController 
{
	_loaded_data = null;
	_loaded_from_version = null;

	_company_tbl = null;

	constructor()
	{
		this._loaded_data = null;
		this._loaded_from_version = null;

		this._company_tbl = {};
	}

	// Return the company data class instance for a company or null if the company doesn't exist
	// in our in-memory list of companies.
	function GetCompanyData(company_id);
}

function MainClass::Start()
{
	this.Init();

	Log.Info("Setup done", Log.LVL_INFO);

	// Wait for the game to start
	this.Sleep(1);

	local last_income_check = GSDate.GetCurrentDate();
	local last_loop_year = GSDate.GetYear(GSDate.GetCurrentDate());
	while (true) {
		local loop_start_tick = GSController.GetTick();

		// Handle events
		this.HandleEvents();

		// Reached end of year?
		local current_date = GSDate.GetCurrentDate();
		local year = GSDate.GetYear(current_date);
		if (year != last_loop_year)
		{
			this.EndOfYear(year - 1);
			last_loop_year = year;
		}
	
		// Once a month
		if (current_date > last_income_check + 30)
		{
			last_income_check = current_date;

			// Add/remove companies in case event was lost
			this.UpdateCompanyList();

			// Update company bank balance based on transported cargo
			GSLog.Info("Scan goals START");
			this.UpdateCompanyBalance();
			GSLog.Info("Scan goals END");
		}

		// Check for incoming SCP messages
		Log.Info("SCP.Check", Log.LVL_SUB_DECISIONS);
		for(local s=0; s<20 && g_scp_manager.Check(); s++) {};

		// Loop with a frequency of five days
		local ticks_used = GSController.GetTick() - loop_start_tick;
		this.Sleep(Helper.Max(1, 5 * 74 - ticks_used));
	}
}

function MainClass::Init()
{
	// Setup SCPManager
	g_scp_manager = SCPManager(this);
	

	if (this._loaded_data != null)
	{
		// Copy loaded data from this._loaded_data to this.*
		// or do whatever with the loaded data

		// Load company data from loaded data
		foreach (company_table in this._loaded_data.company_list)
		{
			if (company_table != null)
			{
				Log.Info("Loading data for company " + GSCompany.GetName(company_table.company_id), Log.LVL_INFO);
				local loaded_company = CompanyData.CreateFromTable(company_table, this._loaded_from_version)
				this._company_tbl.rawset(loaded_company.GetId(), loaded_company);
			}
		}

		this._loaded_data = null; // don't attempt to load again
	}
	else
	{
		// New game
	}

	// Add possible new companies
	this.UpdateCompanyList();
}

function MainClass::HandleEvents()
{
	if (GSEventController.IsEventWaiting())
	{
		local ev = GSEventController.GetNextEvent();

		if (ev == null)
			return;

		local ev_type = ev.GetEventType();
		if (ev_type == GSEvent.ET_COMPANY_NEW ||
				ev_type == GSEvent.ET_COMPANY_BANKRUPT ||
				ev_type == GSEvent.ET_COMPANY_MERGER)
		{
			Log.Info("A company was created/bankrupt/merged => update company list", Log.LVL_INFO);

			// Update the goal list when:
			// - a new company has been created
			// - a company has gone bankrupt
			// - a company has been bought by another company
			this.UpdateCompanyList();
		}
	}
}

function MainClass::UpdateCompanyList()
{
	// Loop over all possible company IDs
	for (local c = GSCompany.COMPANY_FIRST; c <= GSCompany.COMPANY_LAST; c++)
	{
		// Is the company not existing in the game?
		if (GSCompany.ResolveCompanyID(c) == GSCompany.COMPANY_INVALID)
		{
			if (this._company_tbl.rawin(c))
			{
				// Remove data for no longer existing company
				this._company_tbl.rawdelete(c);
			}
			continue;
		}
	
		// If the company can be resolved and exist (CompanyData has been setup) => don't do anything
		if (this._company_tbl.rawin(c)) continue;

		// Company data has not yet been setup for this company
		local company_data = CompanyData(c)
		company_data.InformAboutRules();
		company_data.IncomeUpdate(); // start tracking delivery

		this._company_tbl.rawset(c, company_data);
	}
}

function MainClass::UpdateCompanyBalance()
{
	for (local c = GSCompany.COMPANY_FIRST; c <= GSCompany.COMPANY_LAST; c++)
	{
		if (this._company_tbl.rawin(c)) {
			Log.Info("Update company bank balance for " + GSCompany.GetName(c), Log.LVL_SUB_DECISIONS);
			this._company_tbl.rawget(c).IncomeUpdate();

			// Extra call to handle events to ensure company gone bankrupt is removed and 
			// then possible added again if new company was created quickly. This reduce
			// the risk that you get money that are not yours.
			this.HandleEvents();

			if (this._company_tbl.rawin(c)) this._company_tbl.rawget(c).DoBankTransfer();
		}
	}
}

// Called at end of year (but not when the game end has been reached)
// @param year The year that just ended
function MainClass::EndOfYear(year)
{
	Log.Info("End Of Year", Log.LVL_INFO);

	foreach (_, company_data in this._company_tbl)
	{
		company_data.EndOfYear(year);
	}
}

function MainClass::Save()
{
	Log.Info("Saving data to savegame", Log.LVL_INFO);

	// If Init() has not finished loading data from save,
	// then save the data that was loaded from save.
	if (this._loaded_data != null) return this._loaded_data;

	local company_save_list = [];
	foreach (_, company_data in this._company_tbl)
	{
		company_save_list.append(company_data.SaveToTable());
	}

	return { 
		company_list = company_save_list
	};
}

function MainClass::Load(version, tbl)
{
	Log.Info("Loading data from savegame made with version " + version + " of the game script", Log.LVL_INFO);

	if (version > SELF_VERSION)
	{
		Log.Warning("Warning: Loading from a newer version of TransportGoals", Log.LVL_INFO);
	}

	// Store a copy of the table from the save game
	// but do not process the loaded data yet. Wait with that to Init
	// so that OpenTTD doesn't kick us for taking too long to load.
	this._loaded_data = {}
   	foreach (key, val in tbl)
	{
		this._loaded_data.rawset(key, val);
	}	
	this._loaded_from_version = version;
}

// public function
function MainClass::GetCompanyData(company_id)
{
	return this._company_tbl.rawin(company_id) ? this._company_tbl.rawget(company_id) : null;
}

