
// Global function for checking if SCP is enabled at all
function IsSCPEnabled()
{
	return GSController.GetSetting("scp_enabled") == 1;
}


// SCP Manager class
class SCPManager {
	_main_ptr = null;
	_scp_enabled = null;
	
   static COMMAND_SET = "CashDrainGS";

	constructor(main_ptr)
	{
		this._scp_enabled = IsSCPEnabled();
		this._main_ptr = main_ptr;
		this.SetupSCP();
	}

}

function DataToStr(data)
{
	if (data == null)
		return "NULL";

	local s = "[";
	foreach(d in data)
	{
		if (s != "[") s += ", ";
		if (d == null)
			s += "NULL";
		else if(typeof(d) == "string")
			s += "\"" + d + "\"";
		else
			s += d.tostring();
	}

	return s + "]";
}

function SCPManager::SetupSCP()
{
	if (!this._scp_enabled) return;

	// Initialize the library itself
	local dummy = SCPLib(SELF_SHORTNAME, SELF_VERSION, null); // yes, call constructor and then throw away instance.
	local show_debug = Log.IsLevelAccepted(Log.LVL_DEBUG);
	SCPLib.SCPLogging_Info(show_debug); // <-- static call to library
	SCPLib.SCPLogging_Error(true);

	// Register commands
	local self = this;

	// AI -> GS commands:
	SCPLib.AddCommand("Setting", COMMAND_SET, self, SCPManager.ReceivedSettingCommand);

	// GS -> AI commands:
	//SCPLib.AddCommand("CargoCompleted", COMMAND_SET, self);
}

function SCPManager::Check()
{
	if (!this._scp_enabled) return;

	// Update SCP logging in case the log level was changed
	local show_debug = Log.IsLevelAccepted(Log.LVL_DEBUG);
	SCPLib.SCPLogging_Info(show_debug);

	// Let SCP check for incoming messages 
	return SCPLib.Check();
}

/*****************************************************************
 *                                                               *
 *   Outgoing Commands - commands that we can send to AIs        *
 *                                                               *
 *****************************************************************/

// Call one of these methods to send a command to an AI.


/*****************************************************************
 *                                                               *
 *   Incoming Commands - commands that we can get from AIs       *
 *                                                               *
 *****************************************************************/

// These methods are called by the SCP library when we call SCPLib.Check() and there is
// a received incoming message.

/*
 * Input:
 *   Data[0] => string to pass to GSController.GetSetting(setting)
 *   Data[1] => new value of setting (only accepted by some settings)
 *
 * Output:
 *   Data[0] => setting name string
 *   Data[1] => setting value
 *
 * If the setting doesn't exist, Data[0] will contain the setting name and
 * Data[1] will contain null.
 */
function SCPManager::ReceivedSettingCommand(message, self)
{
	if (!self._scp_enabled) return;

	if(Log.IsLevelAccepted(Log.LVL_SUB_DECISIONS))
	{
		Log.Info("Received Setting Command with data: " + DataToStr(message.Data), Log.LVL_INFO);
	}

	// Get setting string
	local setting = message.GetStringData(0);
	local response_value = null;
	if(setting != null)
	{
		// fake setting?
		if (setting == "version")
		{
			response_value = SELF_VERSION;
		}
		else if(setting == "payment_per_unit") // only white listed GSSetting settings are allowed
		{
			response_value = GSController.GetSetting(setting);
		}
		else
		{
			Log.Info("AI company " + GSCompany.GetName(message.SenderID) + " asked for unknown setting \"" + setting + "\"", Log.LVL_INFO);
		}

		if (response_value != null)
			Log.Info(GSCompany.GetName(message.SenderID) + " asked for setting value of \"" + setting + "\" which has the value: " + response_value, Log.LVL_DEBUG);
	}

	// Answer with the setting value
	SCPLib.Answer(message, setting, response_value);
}
